(module interp (lib "eopl.ss" "eopl")
  
  ;; interpreter for the LET language.  The \commentboxes are the
  ;; latex code for inserting the rules into the code in the book.
  ;; These are too complicated to put here, see the text, sorry.

  (require "drscheme-init.scm")

  (require "lang.scm")
  (require "data-structures.scm")
  (require "environments.scm")

  (provide value-of-program value-of)

;;;;;;;;;;;;;;;; the interpreter ;;;;;;;;;;;;;;;;

  ;; value-of-program : Program -> ExpVal
  ;; Page: 71
  (define value-of-program 
    (lambda (pgm)
      (cases program pgm
        (a-program (exp1)
          (value-of exp1 (init-env))))))

  ;; value-of : Exp * Env -> ExpVal
  ;; Page: 71
  (define value-of
    (lambda (exp env)
      (cases expression exp

        ;\commentbox{ (value-of (const-exp \n{}) \r) = \n{}}
        (const-exp (num) (num-val num))

        ;\commentbox{ (value-of (var-exp \x{}) \r) = (apply-env \r \x{})}
        (var-exp (var) (apply-env env var))

        ;\commentbox{\diffspec}
        (diff-exp (exp1 exp2)
          (let ((val1 (value-of exp1 env))
                (val2 (value-of exp2 env)))
            (let ((num1 (expval->num val1))
                  (num2 (expval->num val2)))
              (num-val
                (- num1 num2)))))

        ;\commentbox{\zerotestspec}
        (zero?-exp (exp1)
          (let ((val1 (value-of exp1 env)))
            (let ((num1 (expval->num val1)))
              (if (zero? num1)
                (bool-val #t)
                (bool-val #f)))))
              
        ;\commentbox{\ma{\theifspec}}
        (if-exp (exp1 exp2 exp3)
          (let ((val1 (value-of exp1 env)))
            (if (expval->bool val1)
              (value-of exp2 env)
              (value-of exp3 env))))

        ;\commentbox{\ma{\theletspecsplit}}
        (let-exp (var exp1 body)       
          (let ((val1 (value-of exp1 env)))
            (value-of body
              (extend-env var val1 env))))

        ;commentbox{\minusspec}
        (minus-exp (exp) (num-val (- 0 (expval->num (value-of exp env)))))
        
        ;\commentbox{\addspec}
        (add-exp (exp1 exp2)
          (let ((val1 (value-of exp1 env))
                (val2 (value-of exp2 env)))
            (let ((num1 (expval->num val1))
                  (num2 (expval->num val2)))
              (num-val
                (+ num1 num2)))))
        
        ;\commentbox{\multspec}
        (mult-exp (exp1 exp2)
          (let ((val1 (value-of exp1 env))
                (val2 (value-of exp2 env)))
            (let ((num1 (expval->num val1))
                  (num2 (expval->num val2)))
              (num-val
                (* num1 num2)))))
        
        ;\commentbox{\quotspec}
        (quot-exp (exp1 exp2)
          (let ((val1 (value-of exp1 env))
                (val2 (value-of exp2 env)))
            (let ((num1 (expval->num val1))
                  (num2 (expval->num val2)))
              (num-val
                (quotient num1 num2)))))
        
        ;\commentbox{\equalspec}
        (equal-exp (exp1 exp2)
          (let ((val1 (value-of exp1 env))
                (val2 (value-of exp2 env)))
            (let ((num1 (expval->num val1))
                  (num2 (expval->num val2)))
              (bool-val
                (equal? num1 num2)))))
        
        ;\commentbox{\greaterspec}
        (greater-exp (exp1 exp2)
          (let ((val1 (value-of exp1 env))
                (val2 (value-of exp2 env)))
            (let ((num1 (expval->num val1))
                  (num2 (expval->num val2)))
              (bool-val
                (> num1 num2)))))
        
        ;\commentbox{\lessspec}
        (less-exp (exp1 exp2)
          (let ((val1 (value-of exp1 env))
                (val2 (value-of exp2 env)))
            (let ((num1 (expval->num val1))
                  (num2 (expval->num val2)))
              (bool-val
                (< num1 num2)))))
        
        ;\commentbox{\consspec}
        (cons-exp (exp1 exp2)
           (pair-val (value-of exp1 env) (value-of exp2 env)))
        
        ;\commentbox{\nullspec}
        (null-exp ()
           (null-val))
        
        ;\commentbox{\carspec}
        (car-exp (exp)
           (expval->lhs (value-of exp env)))
        
        ;\commentbox{\cdrspec}
        (cdr-exp (exp)
           (expval->rhs (value-of exp env)))
        
        ;\commentbox{\null?spec}
        (null?-exp (exp)
            (cases expval (value-of exp env) 
              (null-val () (bool-val #t))
              (else (bool-val #f))))
        
        ;\commentbox{\condspec}
        (cond-exp (cexp)
           (value-of-cond-exp cexp env)
         )
        
        ;\commentbox{\unpackspec}
        (unpack-exp (ids lst exp)
           (value-of exp (extend-env-ids ids (value-of lst env) env))      
         )
        
        )))
  
  ;; extend-env-ids : MultId * Exp * Env -> ExpVal
  (define (extend-env-ids ids lst env)
    (cases mult-id ids
      (empty-ids () (if (equal? lst (null-val)) env (eopl:error "unpack error")))
      (rec-ids (id mid) 
               (if (not (equal? lst (null-val))) 
                            (extend-env-ids mid (expval->rhs lst)
                                            (extend-env id (expval->lhs lst) env))
                            (eopl:error "unpack error")))))

  ;; value-of-cond-exp : CondExp * Env -> ExpVal
  (define value-of-cond-exp
    (lambda (exp env)
      (cases cond-expression exp
        (rec-cond-exp (lexp rexp cexp)
           (cond
             [(equal? (value-of lexp env) (bool-val #t)) (value-of rexp env)]
             [else (value-of-cond-exp cexp env)]
             ))
        (empty-cond-exp ()
          (eopl:error "invalid condition")))))


  )

